/* strtoull - convert string representation of a number into an unsigned long long value. */

/* Copyright (C) 1997 Free Software Foundation, Inc.

   This file is part of GNU Bash, the Bourne Again SHell.

   Bash is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Bash is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Bash.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <config.h>

#if defined (HAVE_LONG_LONG) && !defined (HAVE_STRTOULL)

#define QUAD		1
#define	UNSIGNED	1
#undef HAVE_STRTOL

#include "strtol.c"

#endif /* HAVE_LONG_LONG && !HAVE_STRTOULL */
